"""SSH client for NanoKVM terminal access."""

from __future__ import annotations

import asyncio

import paramiko

from .client import NanoKVMError

DEFAULT_SSH_USERNAME = "root"


class NanoKVMSSHError(NanoKVMError):
    """Base exception for SSH client errors."""


class NanoKVMSSHNotConnectedError(NanoKVMSSHError):
    """Exception for when SSH client is not connected."""


class NanoKVMSSHAuthenticationError(NanoKVMSSHError):
    """Exception for SSH authentication failures."""


class NanoKVMSSHCommandError(NanoKVMSSHError):
    """Exception for SSH command execution errors."""


class NanoKVMSSH:
    """SSH client for NanoKVM terminal access."""

    def __init__(
        self, host: str, username: str = DEFAULT_SSH_USERNAME, port: int = 22
    ) -> None:
        """Initialize the SSH client."""
        self.host = host
        self.port = port
        self.username = username
        self.ssh_client: paramiko.SSHClient | None = None

    async def authenticate(self, password: str) -> None:
        """Authenticate with SSH using password."""
        loop = asyncio.get_running_loop()
        self.ssh_client = paramiko.SSHClient()
        self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        try:
            client = self.ssh_client  # Capture for lambda
            await loop.run_in_executor(
                None,
                lambda: client.connect(
                    self.host,
                    port=self.port,
                    username=self.username,
                    password=password,
                    timeout=10,
                ),
            )
        except paramiko.AuthenticationException as e:
            raise NanoKVMSSHAuthenticationError(
                f"SSH authentication failed: {e}"
            ) from e
        except (paramiko.SSHException, paramiko.BadHostKeyException, OSError) as e:
            raise NanoKVMSSHAuthenticationError(f"SSH connection failed: {e}") from e

    async def disconnect(self) -> None:
        """Close SSH connection."""
        if self.ssh_client:
            self.ssh_client.close()
            self.ssh_client = None

    async def run_command(self, command: str, timeout: int = 30) -> str:
        """Run a command via SSH and return output."""
        if not self.ssh_client:
            raise NanoKVMSSHNotConnectedError(
                "SSH not connected, call authenticate first"
            )
        loop = asyncio.get_running_loop()
        try:
            output, error = await asyncio.wait_for(
                loop.run_in_executor(None, self._exec_command_sync, command),
                timeout=timeout,
            )
            if error:
                raise NanoKVMSSHCommandError(f"SSH command error: {error}")
            return output.strip()
        except asyncio.TimeoutError:
            raise NanoKVMSSHCommandError(
                f"SSH command timed out after {timeout} seconds"
            ) from None

    def _exec_command_sync(self, command: str) -> tuple[str, str]:
        """Synchronous SSH command execution."""
        assert self.ssh_client is not None  # Should be set after authenticate()
        stdin, stdout, stderr = self.ssh_client.exec_command(command)
        output = stdout.read().decode("utf-8")
        error = stderr.read().decode("utf-8")
        return output, error



from aioresponses import aioresponses
import pytest

from nanokvm.client import NanoKVMApiError, NanoKVMClient
from nanokvm.models import ApiResponseCode


async def test_get_images_success() -> None:
    """Test get_images with a successful response."""
    async with NanoKVMClient(
        "http://localhost:8888/api/", token="test-token"
    ) as client:
        with aioresponses() as m:
            m.get(
                "http://localhost:8888/api/storage/image",
                payload={
                    "code": 0,
                    "msg": "success",
                    "data": {
                        "files": [
                            "/data/alpine-standard-3.23.2-x86_64.iso",
                            "/data/cs10-js.iso",
                        ]
                    },
                },
            )

            response = await client.get_images()

            assert response is not None
            assert len(response.files) == 2
            assert "/data/alpine-standard-3.23.2-x86_64.iso" in response.files
            assert "/data/cs10-js.iso" in response.files


async def test_get_images_empty() -> None:
    """Test get_images with an empty list."""
    async with NanoKVMClient(
        "http://localhost:8888/api/", token="test-token"
    ) as client:
        with aioresponses() as m:
            m.get(
                "http://localhost:8888/api/storage/image",
                payload={"code": 0, "msg": "success", "data": {"files": []}},
            )

            response = await client.get_images()

            assert response is not None
            assert len(response.files) == 0


async def test_get_images_api_error() -> None:
    """Test get_images with an API error response."""
    async with NanoKVMClient(
        "http://localhost:8888/api/", token="test-token"
    ) as client:
        with aioresponses() as m:
            m.get(
                "http://localhost:8888/api/storage/image",
                payload={"code": -1, "msg": "failed to list images", "data": None},
            )

            with pytest.raises(NanoKVMApiError) as exc_info:
                await client.get_images()

            assert exc_info.value.code == ApiResponseCode.FAILURE
            assert "failed to list images" in exc_info.value.msg


async def test_client_context_manager() -> None:
    """Test that client properly initializes and cleans up with context manager."""
    async with NanoKVMClient(
        "http://localhost:8888/api/", token="test-token"
    ) as client:
        # Verify session is created
        assert client._session is not None
        assert not client._session.closed

    # After exiting context, session should be closed
    assert client._session is None



"""Tests for mouse control functionality."""

from collections.abc import AsyncGenerator
from typing import Any
from unittest.mock import AsyncMock, patch

import pytest

from nanokvm.client import NanoKVMClient
from nanokvm.models import MouseButton


@pytest.fixture
async def client_with_mock_ws() -> AsyncGenerator[tuple[NanoKVMClient, AsyncMock], Any]:
    """Fixture that provides a client with mocked WebSocket."""
    mock_ws = AsyncMock()
    mock_ws.closed = False

    with patch(
        "aiohttp.ClientSession.ws_connect", new_callable=AsyncMock, return_value=mock_ws
    ):
        async with NanoKVMClient(
            "http://localhost:8888/api/", token="test-token"
        ) as client:
            yield client, mock_ws


async def test_mouse_move_abs(
    client_with_mock_ws: tuple[NanoKVMClient, AsyncMock],
) -> None:
    """Test absolute mouse movement."""
    client, mock_ws = client_with_mock_ws

    await client.mouse_move_abs(0.5, 0.5)

    # Verify the WebSocket send was called with correct parameters
    mock_ws.send_json.assert_called_once()
    message = mock_ws.send_json.call_args[0][0]
    # Message format: [2, event_type, button_state, x_val, y_val]
    assert message[0] == 2  # mouse event indicator
    assert message[1] == 2  # move_abs
    assert message[2] == 0  # button state
    # 0.5 * 32768 = 16384
    assert message[3] == 16384
    assert message[4] == 16384


async def test_mouse_move_rel(
    client_with_mock_ws: tuple[NanoKVMClient, AsyncMock],
) -> None:
    """Test relative mouse movement."""
    client, mock_ws = client_with_mock_ws

    await client.mouse_move_rel(0.1, -0.1)

    # Verify the WebSocket send was called with correct parameters
    mock_ws.send_json.assert_called_once()
    message = mock_ws.send_json.call_args[0][0]
    # Message format: [2, event_type, button_state, x_val, y_val]
    assert message[0] == 2  # mouse event indicator
    assert message[1] == 3  # move_rel
    assert message[2] == 0  # button state
    # 0.1 * 32768 = 3276.8 -> 3276
    assert message[3] == 3276
    # -0.1 * 32768 = -3276.8 -> -3276
    assert message[4] == -3276


async def test_mouse_down(client_with_mock_ws: tuple[NanoKVMClient, AsyncMock]) -> None:
    """Test mouse button down."""
    client, mock_ws = client_with_mock_ws

    await client.mouse_down(MouseButton.LEFT)

    # Verify the WebSocket send was called with correct parameters
    mock_ws.send_json.assert_called_once()
    message = mock_ws.send_json.call_args[0][0]
    # Message format: [2, event_type, button_state, x_val, y_val]
    assert message[0] == 2  # mouse event indicator
    assert message[1] == 1  # mouse_down
    assert message[2] == 1  # left button
    assert message[3] == 0
    assert message[4] == 0


async def test_mouse_up(client_with_mock_ws: tuple[NanoKVMClient, AsyncMock]) -> None:
    """Test mouse button up."""
    client, mock_ws = client_with_mock_ws

    await client.mouse_up()

    # Verify the WebSocket send was called with correct parameters
    mock_ws.send_json.assert_called_once()
    message = mock_ws.send_json.call_args[0][0]
    # Message format: [2, event_type, button_state, x_val, y_val]
    assert message[0] == 2  # mouse event indicator
    assert message[1] == 0  # mouse_up
    assert message[2] == 0  # button state (always 0 for mouse_up)
    assert message[3] == 0
    assert message[4] == 0


async def test_mouse_click_without_position(
    client_with_mock_ws: tuple[NanoKVMClient, AsyncMock],
) -> None:
    """Test mouse click at current position."""
    client, mock_ws = client_with_mock_ws

    await client.mouse_click(MouseButton.LEFT)

    # Should send two events: down and up
    assert mock_ws.send_json.call_count == 2

    # First call should be mouse_down
    first_message = mock_ws.send_json.call_args_list[0][0][0]
    assert first_message[0] == 2  # mouse event indicator
    assert first_message[1] == 1  # mouse_down
    assert first_message[2] == 1  # left button

    # Second call should be mouse_up
    second_message = mock_ws.send_json.call_args_list[1][0][0]
    assert second_message[0] == 2  # mouse event indicator
    assert second_message[1] == 0  # mouse_up
    assert second_message[2] == 0  # button state (always 0 for mouse_up)


async def test_mouse_click_with_position(
    client_with_mock_ws: tuple[NanoKVMClient, AsyncMock],
) -> None:
    """Test mouse click at specific position."""
    client, mock_ws = client_with_mock_ws

    await client.mouse_click(MouseButton.MIDDLE, 0.25, 0.75)

    # Should send three events: move_abs, down, and up
    assert mock_ws.send_json.call_count == 3

    # First call should be move_abs
    first_message = mock_ws.send_json.call_args_list[0][0][0]
    assert first_message[0] == 2  # mouse event indicator
    assert first_message[1] == 2  # move_abs
    assert first_message[3] == int(0.25 * 32768)
    assert first_message[4] == int(0.75 * 32768)

    # Second call should be mouse_down
    second_message = mock_ws.send_json.call_args_list[1][0][0]
    assert second_message[0] == 2  # mouse event indicator
    assert second_message[1] == 1  # mouse_down
    assert second_message[2] == 4  # middle button

    # Third call should be mouse_up
    third_message = mock_ws.send_json.call_args_list[2][0][0]
    assert third_message[0] == 2  # mouse event indicator
    assert third_message[1] == 0  # mouse_up
    assert third_message[2] == 0  # button state (always 0 for mouse_up)


async def test_mouse_scroll(
    client_with_mock_ws: tuple[NanoKVMClient, AsyncMock],
) -> None:
    """Test mouse scroll."""
    client, mock_ws = client_with_mock_ws

    await client.mouse_scroll(0.1, -0.2)

    # Verify the WebSocket send was called with correct parameters
    mock_ws.send_json.assert_called_once()
    message = mock_ws.send_json.call_args[0][0]
    # Message format: [2, event_type, button_state, x_val, y_val]
    assert message[0] == 2  # mouse event indicator
    assert message[1] == 4  # scroll
    assert message[2] == 0  # button state
    # 0.1 * 32768 = 3276.8 -> 3276
    assert message[3] == 3276
    # -0.2 * 32768 = -6553.6 -> -6553
    assert message[4] == -6553


async def test_context_manager_cleanup() -> None:
    """Test that context manager properly closes WebSocket and session."""
    mock_ws = AsyncMock()
    mock_ws.closed = False

    with patch(
        "aiohttp.ClientSession.ws_connect", new_callable=AsyncMock, return_value=mock_ws
    ):
        async with NanoKVMClient(
            "http://localhost:8888/api/", token="test-token"
        ) as client:
            # Trigger WebSocket creation by making a call
            await client.mouse_move_abs(0.0, 0.0)
            # Verify WebSocket was created
            assert client._ws is not None
            assert client._session is not None

        # After exiting context, resources should be cleaned up
        mock_ws.close.assert_called_once()
        assert client._ws is None
        assert client._session is None


async def test_button_mapping(
    client_with_mock_ws: tuple[NanoKVMClient, AsyncMock],
) -> None:
    """Test different button types."""
    client, mock_ws = client_with_mock_ws

    # Test left button
    await client.mouse_down(MouseButton.LEFT)
    message = mock_ws.send_json.call_args[0][0]
    assert message[2] == MouseButton.LEFT

    # Test right button
    await client.mouse_down(MouseButton.RIGHT)
    message = mock_ws.send_json.call_args[0][0]
    assert message[2] == MouseButton.RIGHT

    # Test middle button
    await client.mouse_down(MouseButton.MIDDLE)
    message = mock_ws.send_json.call_args[0][0]
    assert message[2] == MouseButton.MIDDLE



import base64
import hashlib
import os
import urllib.parse

from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

# XXX: this offers no security whatsoever, since this is a symmetric cipher with a
# PUBLIC key. This is no more or less secure than ROT13.
SECRET_KEY = b"nanokvm-sipeed-2024"


def evp_bytes_to_key_aes256_md5(
    password: bytes, salt: bytes, *, key_len: int = 32, iv_len: int = 16
) -> tuple[bytes, bytes]:
    """OpenSSL's `EVP_BytesToKey` function with a few hardcoded parameters."""
    derived = b""
    block = b""

    while len(derived) < key_len + iv_len:
        block = hashlib.md5(block + password + salt).digest()
        derived += block

    key = derived[0:key_len]
    iv = derived[key_len : key_len + iv_len]

    return key, iv


def openssl_encrypt_aes256cbc_md5(plaintext: bytes, password: bytes) -> bytes:
    """Implementation of `CryptoJS.AES.encrypt`."""

    salt = os.urandom(8)
    key, iv = evp_bytes_to_key_aes256_md5(password, salt)

    padder = padding.PKCS7(128).padder()
    padded_plaintext = padder.update(plaintext) + padder.finalize()

    cipher = Cipher(algorithms.AES256(key), modes.CBC(iv))
    encryptor = cipher.encryptor()
    ciphertext: bytes = encryptor.update(padded_plaintext) + encryptor.finalize()

    return b"Salted__" + salt + ciphertext


def obfuscate_password(password: str) -> str:
    """Obfuscate a password."""
    password_enc = openssl_encrypt_aes256cbc_md5(
        plaintext=password.encode("utf-8"),
        password=SECRET_KEY,
    )

    return urllib.parse.quote(base64.b64encode(password_enc).decode("utf-8"), safe="")



"""Models for NanoKVM API."""

from __future__ import annotations

from enum import IntEnum, StrEnum
from typing import Generic, TypeVar

from pydantic import BaseModel, Field

T = TypeVar("T")


class ApiResponseCode(IntEnum):
    """API Response Codes."""

    SUCCESS = 0
    FAILURE = -1
    INVALID_USERNAME_OR_PASSWORD = -2


class HidMode(StrEnum):
    """HID Operating Modes."""

    NORMAL = "normal"
    HID_ONLY = "hid-only"


class GpioType(StrEnum):
    """GPIO Control types."""

    RESET = "reset"
    POWER = "power"


class ScreenSettingType(StrEnum):
    """Screen Setting types."""

    RESOLUTION = "resolution"
    FPS = "fps"
    QUALITY = "quality"


class RunScriptType(StrEnum):
    """Script Execution types."""

    FOREGROUND = "foreground"
    BACKGROUND = "background"


class VirtualDevice(StrEnum):
    """Virtual Device types."""

    NETWORK = "network"
    DISK = "disk"


class TailscaleState(StrEnum):
    """Tailscale Service states."""

    NOT_INSTALLED = "notInstall"
    NOT_RUNNING = "notRunning"
    NOT_LOGIN = "notLogin"
    STOPPED = "stopped"
    RUNNING = "running"


class TailscaleAction(StrEnum):
    """Tailscale Service action."""

    INSTALL = "install"
    UNINSTALL = "uninstall"
    UP = "up"
    DOWN = "down"
    LOGIN = "login"
    LOGOUT = "logout"
    START = "start"
    STOP = "stop"
    RESTART = "restart"


class DownloadStatus(StrEnum):
    """Download Status."""

    IDLE = "idle"
    IN_PROGRESS = "in_progress"


class HWVersion(StrEnum):
    """Hardware Version Enum based on Go constants."""

    ALPHA = "Alpha"
    BETA = "Beta"
    PCIE = "PCIE"
    PRO = "Pro"
    UNKNOWN = "Unknown"


class MouseJigglerMode(StrEnum):
    """Mouse Jiggler Modes."""

    ABSOLUTE = "absolute"
    RELATIVE = "relative"


class MouseButton(IntEnum):
    """Mouse Button types."""

    LEFT = 1
    RIGHT = 2
    MIDDLE = 4


# Generic Response Wrapper
class ApiResponse(BaseModel, Generic[T]):
    """Generic API response structure."""

    code: ApiResponseCode
    msg: str
    data: T | None = None


# Authentication Models
class LoginReq(BaseModel):
    username: str
    password: str


class LoginRsp(BaseModel):
    token: str


class GetAccountRsp(BaseModel):
    username: str


class ChangePasswordReq(BaseModel):
    username: str
    password: str


class IsPasswordUpdatedRsp(BaseModel):
    is_updated: bool = Field(alias="isUpdated")


# VM Models
class IPInfo(BaseModel):
    """IP Address Information."""

    name: str
    addr: str
    version: str
    type: str


class GetInfoRsp(BaseModel):
    ips: list[IPInfo]
    mdns: str
    image: str
    application: str
    device_key: str = Field(alias="deviceKey")


class GetHardwareRsp(BaseModel):
    version: HWVersion


class SetGpioReq(BaseModel):
    type: GpioType
    duration: int  # Milliseconds


class GetGpioRsp(BaseModel):
    pwr: bool  # Power LED state
    hdd: bool  # HDD LED state (only valid for Alpha hardware)


class SetScreenReq(BaseModel):
    type: ScreenSettingType
    value: int


class GetVirtualDeviceRsp(BaseModel):
    network: bool
    disk: bool


class UpdateVirtualDeviceReq(BaseModel):
    device: VirtualDevice


class UpdateVirtualDeviceRsp(BaseModel):
    on: bool


class GetMemoryLimitRsp(BaseModel):
    enabled: bool
    limit: int  # In MB


class SetMemoryLimitReq(BaseModel):
    enabled: bool
    limit: int  # In MB


class GetOLEDRsp(BaseModel):
    exist: bool
    sleep: int  # Sleep timeout in seconds


class SetOledReq(BaseModel):
    sleep: int  # Sleep timeout in seconds


class GetSSHStateRsp(BaseModel):
    enabled: bool


class GetSwapSizeRsp(BaseModel):
    size: int


class SetSwapSizeReq(BaseModel):
    size: int


class GetMdnsStateRsp(BaseModel):
    enabled: bool


# HID Models
class GetHidModeRsp(BaseModel):
    mode: HidMode


class SetHidModeReq(BaseModel):
    mode: HidMode


class PasteReq(BaseModel):
    content: str


# Storage Models
class GetImagesRsp(BaseModel):
    files: list[str]


class MountImageReq(BaseModel):
    file: str | None = None
    cdrom: bool | None = None


class GetMountedImageRsp(BaseModel):
    file: str  # Path to the mounted file, empty if none or default


class GetCdRomRsp(BaseModel):
    cdrom: int


# Network Models
class WakeOnLANReq(BaseModel):
    mac: str


class GetMacRsp(BaseModel):
    macs: list[str]


class DeleteMacReq(BaseModel):
    mac: str


class GetWifiRsp(BaseModel):
    supported: bool
    connected: bool


class ConnectWifiReq(BaseModel):
    ssid: str
    password: str  # Plain text password


class GetTailscaleStatusRsp(BaseModel):
    state: TailscaleState
    name: str
    ip: str
    account: str


# Application Models
class GetVersionRsp(BaseModel):
    current: str
    latest: str


class GetPreviewRsp(BaseModel):
    enabled: bool


class SetPreviewReq(BaseModel):
    enable: bool


# Download Models
class ImageEnabledRsp(BaseModel):
    enabled: bool


class StatusImageRsp(BaseModel):
    status: DownloadStatus
    file: str
    percentage: str


class DownloadImageReq(BaseModel):
    file: str  # URL of the image to download
    # cdrom field is ignored for downloads


class SetMouseJigglerReq(BaseModel):
    enabled: bool
    mode: MouseJigglerMode


class GetMouseJigglerRsp(BaseModel):
    enabled: bool
    mode: MouseJigglerMode


class GetHdmiStateRsp(BaseModel):
    enabled: bool
